#include <u.h>
#include <libc.h>

enum {
	Suser,
	Snice,
	Ssystem,
	Sidle,
	Siowait,
	Sirq,
	Ssoftirq,
	Nstats,

	/* 4GB io default */
	Dbs = 1<<20,
	Dcount = 1<<12,

	USER_HZ = 100,
};

char *snames[] = {
	[Suser] = "user",
	[Snice] = "nice",
	[Ssystem] = "system",
	[Sidle] = "idle",
	[Siowait] = "iowait",
	[Sirq] = "irq",
	[Ssoftirq] = "softirq"
};

ulong rdstats[Nstats];
ulong wrstats[Nstats];
char file[4096];
char *block;
int fd;
vlong bs = Dbs;
vlong count = Dcount;
int fflag, ncpu;

void
procstat(ulong stats[Nstats])
{
	int i, n, fd, argc;
	char buf[128], *p, *argv[16], **args;

	args = argv;
	fd = open("/proc/stat", O_RDONLY);
	if (fd < 0)
		sysfatal("procstat open failure: %r.");
	n = read(fd, buf, sizeof buf - 1);
	if (n < 0)
		sysfatal("procstat read failure: %r.");
	buf[n] = '\0';
	p = strchr(buf, '\n');
	if (p == nil)
		sysfatal("procstat can't find eol in '%s'.", buf);
	*p = '\0';
	argc = tokenize(buf, argv, nelem(argv));
	if (argc < Nstats + 1)
		sysfatal("procstat format assumption failure.");
	args++;
	for (i=0; i<Nstats; i++)
		stats[i] = atoi(args[i]) - stats[i];
	close(fd);
}

void
ddr(void)
{
	int i, n;

	if (lseek(fd, 0, SEEK_SET) == -1)
		sysfatal("unable to seek to zero for read.");

	print("Reading from %s ... ", file);
	procstat(rdstats);
	for (i=0; i<count; i++) {
		n = read(fd, block, bs);
		if (n != bs)
			sysfatal("short read %d, block %d", n, i);
	}
	procstat(rdstats);
	print("done.\n");
}

void
ddw(void)
{
	int i, n;

	if (lseek(fd, 0, SEEK_SET) == -1)
		sysfatal("unable to seek to zero for write.");

	print("Writing to %s ... ", file);
	procstat(wrstats);
	for (i=0; i<count; i++) {
		n = write(fd, block, bs);
		if (n != bs)
			sysfatal("short write %d, block %d", n, i);
	}
	print("syncing ... ");
	sync();
	procstat(wrstats);
	print("done.\n");
}

void
raw(ulong stats[Nstats])
{
	int i;

	for (i=0; i<Nstats; i++)
		print("%s=%d ", snames[i], stats[i]);
	print("\n");
}

void
cooked(ulong stats[Nstats])
{
	vlong kbs, m, n;
	ulong cpuu;
	int i;

	for (i=0, n=100; i<Nstats; i++)
		n += stats[i];
	kbs = (bs * count);
	kbs /= n;
	kbs *= USER_HZ * ncpu;
	kbs >>= 10;

	m = stats[Sidle] + stats[Siowait];
	cpuu = (n - m) * 100 / n;

	print("%#6lld %#6ld\n", kbs, cpuu);
}

void
ddt(void)
{
	print("%-#5lldMB  KiB/s    CPU%% \n", (bs*count) >> 20);
	print("Write   ");
	cooked(wrstats);
	print("Read    ");
	cooked(rdstats);
	print("\n");
	print("WRaw ");
	raw(wrstats);
	print("RRaw ");
	raw(rdstats);
}

int
countcpus(void)
{
	char buf[4096], *p, *r;
	int fd, n = -1;

	fd = open("/proc/stat", OREAD);
	if (fd < 0)
		sysfatal("cpus: can't open /proc/stat");
	read(fd, buf, sizeof buf - 1);
	buf[sizeof buf - 1] = '\0';
	for (p=r=buf; r=strchr(p, '\n'); p=r) {
		*r++ = '\0';
		if (strncmp(p, "cpu", 3) == 0)
			n++;
	}
	return n;
}

void
main(int argc, char *argv[])
{
	char *dir;
	int n;

	ARGBEGIN {
	case 'b':
		bs = atoi(ARGF());
		break;
	case 'c':
		count = atoi(ARGF());
		break;
	case 'f':
		strcpy(file, ARGF());
		fflag++;
		break;
	case '?':
		goto usage;
	} ARGEND

	if (argc != 1) {
usage:		fprint(2, "usage: %s [-?] [-c count] [-b bs] dir\n", argv0); 
		exits("usage");
	}

	if (fflag == 0) {
		dir = *argv;
		n = strlen(dir);
		if (dir[--n] == '/')
			dir[n] = '\0';
		sprint(file, "%s/ddt.%d", dir, getpid());
		fd = create(file, ORDWR|ORCLOSE, 0777);
	} else
		fd = open(file, ORDWR);

	if (fd == -1)
		sysfatal("can't open %s: %r.", file);

	ncpu = countcpus();
	block = malloc(bs);
	if (block == nil)
		sysfatal("can't malloc %d size block for I/O.", bs);
	ddw();
	print("sleeping 10 seconds ... ");
	sleep(10*1000);
	print("done.\n");
	ddr();
	ddt();
	exits(nil);
}


